/* -*- mode: c++; c-basic-offset: 4; -*- */
//
//    Designer dataset writer version 1.2
//
// debug levels
// 0 No debug output
// 1 Print filenames written, configuration messages.
// 2 file operations (open/close)
// 3 Enter/exit internal functions.
// 4 Channel operations.
//
#include "FrWrite.hh"
#include "FrameSize.hh"
#include "Dacc.hh"
#include "frame_name.hh"
#include "ParseLine.hh"
#include "OperStateCondList.hh"
#include "framecpp/FrameH.hh"
#include "framecpp/FrAdcData.hh"
// #include "framecpp/Dimension.hh"
#include "framecpp/FrVect.hh"

#include "framecpp/GPSTime.hh"
using FrameCPP::GPSTime;

#define GENERAL_FSTREAM 1
#include <fstream>

#include "framecpp/Common/IOStream.hh"
using FrameCPP::Common::FrameBuffer;
#ifdef GENERAL_FSTREAM
#include <ldastoolsal/fstream.hh>
typedef LDASTools::AL::filebuf filebuf_type;
#else
typedef std::filebuf filebuf_type;
#endif
typedef FrameCPP::Common::FrameBuffer< filebuf_type > frame_buffer;

#include <cstring>
#include <cstdio>
#include <algorithm>

using namespace std;
using FrameCPP::FrameH;
using FrameCPP::FrRawData;
using FrameCPP::FrAdcData;
using FrameCPP::Dimension;
using FrameCPP::FrVect;

//======================================  Generate the main function.
EXECDMT(FrWrite)

//======================================  Translate from General::GPSTime
inline Time
GetGPSTime(const GPSTime& T) {
    return Time(T.GetSeconds(), T.GetNanoseconds());
}

inline GPSTime
GetGPSTime(const Time& T) {
    return GPSTime(T.getS(), T.getN());
}


//======================================  Constructor
FrWrite::FrWrite(int argc, const char *argv[]) 
  : DMTBase(argc, argv), mFrOut(1), mFrDone(0), mFrTotal(0), mFileOut(0),
    mRepeat(1), mSequence(0), mMode(m_CopyAll), mResize(0), mTOC(false), 
    mFrameBuf(0), mOSC(0), mTrigName("Trigger"), mBackFrame(0), mTrigFrame(1),
    mBackWrite(0), mTimeOff(0)
{
    const char* synText[] = {
      "FrWrite copies DMT online frame data to a file.",
      " ",
      "Syntax: ",
      "FrWrite [-ofile <file>] [-nframe <nframe>] [-cfile <conf>]",
      "        [-repeat <nfiles>] [{--start-time | -st} {+ -}<start>]",
      "        [+toc] [<dmt args>]",
      "<file> specifies an output file name template and may include any",
      "TimeStr() escape sequence. If <file> is not specified,",
      "'FrWrite-%Y.%m.%d-%H.%N.%S.F'is used. <nframe> frames are written to",
      "each of <nfiles> files. <nframe> and <nfiles> default to 1. If ",
      "'-repeat 0' is specified FrWrite will write files continuously until ",
      "it is killed. <conf> names a configuration file listing channels to ",
      "be written. If unspecified, all channels are included. If +toc is ",
      "specified, a table of contents will be used to access channels data. ",
      "<start> didcated the starting GPS time. If the argument is preceded by ",
      "a sign symbol (+ or -) it is taken to be a time difference in seconds ",
      "between the input and output frame time stamps.",
      0
    };

    const char* file = "FrWrite-%Y.%02m.%02d-%02H.%02N.%02S.F";
    const char* cfile = 0;
    bool ra = false;

    //--------------------------------  Parse the arguments
    for (int i=1 ; i<argc ; i++) {
        string argi = argv[i];
        if (argi == "-ofile") {
	    file = argv[++i];
        } else if (argi == "-cfile") {
	    cfile = argv[++i];
        } else if (argi == "-nframe") {
	    mFrOut = strtol(argv[++i], 0, 0);
        } else if (argi == "-repeat") {
	    mRepeat = strtol(argv[++i], 0, 0);
        } else if (argi == "-resize") {
	    mResize = strtod(argv[++i], 0);
        } else if (argi == "-sequence") {
	    mSequence = strtol(argv[++i], 0, 0);
        } else if (argi == "-toc") {
	    mTOC = false;
        } else if (argi == "+toc") {
	    mTOC = true;
        } else if (argi == "+readall" || argi == "+ra") {
	    ra = true;
        }
	else if (argi == "-backframe" || argi == "-bf") {
	    mBackFrame = strtol(argv[++i], 0, 0);
        }

	//------------------------------  Get the start time or offset.
	else if (argi == "--start-time" || argi == "-st") {
	    i++;
	    if (*argv[i] == '+' || *argv[i] == '-') {
		mTimeOff = strtod(argv[i],0);
	    } else {
		mStartTime = Time(strtol(argv[i], 0, 0));
	    }
        }
	else if (argi == "-trigframe" || argi == "-tf") {
	    mTrigFrame = strtol(argv[++i], 0, 0);
        }
	else if (argi == "-oscfile") {
	    mOSC = new OperStateCondList(getDacc());
	    mOSC->readConfig(argv[++i]);
	}
	else if (argi == "-trigger") {
	    mTrigName = argv[++i];
        } else if (isEnvArg(argi.c_str())) {
	    i++;
	} else {
	    cerr << "Unrecognized argument: " << argi << endl;
	    for (const char** p=synText ; *p ; p++) cerr << *p << endl;
	    cerr << endl;
	    finish();
	    return;
	}
    }
    
    if (mFrOut <= 0) mFrOut = 1;
    if (mOSC && mOSC->find(mTrigName) == mOSC->end()) {
	cerr << "No valid definition of '" << mTrigName 
	     << "' condition." << endl;
	finish();
	return;
    }

    //--------------------------------  Read the configuration file.
    ifstream cstream;
    if (!cfile) {
        cerr << "No configuration file specified. Copy all channels." << endl;
	mMode = m_CopyAll;
    } else {
        cstream.open(cfile, ios::in);
        if (cstream.bad()) {
            cerr << "Error opening configuration file " << cfile << endl;
	    finish();
	    return;
	} else {
            readConfig(cstream);
	    mMode = m_Select;
	    cstream.close();
	}
    }

    //--------------------------------  Set up data accessor
    setHdrOnly(mTOC);
    if (mMode == m_CopyAll && mTOC) {
	cerr << "*** Warning: Copy-all with TOC mode may result in "
	     << "writing empty frames" << endl;
    }

    //--------------------------------  Set up based on parameters
    mOutFile = file;
    if (ra) setBuffer(-1);
    if (!mOSC) {
        mBackFrame = 0;
	mTrigFrame = 1;
    } else if (mBackFrame > 16) {
        cerr << "Backlog reset to maximum (16)." << endl;
        mBackFrame = 16;
    }
}

//======================================  Read the configuration file.
bool
FrWrite::readConfig(istream& fd) {
    ParseLine pl(fd);
    if (!pl.isOpen()) return true;

    //--------------------------------  Read in the configuration file
    while (pl.getLine() >= 0) {
        if (pl.getCount() > 0) mChannel.push_back(ChanDesc(pl[0]));
    }
    if (Debug()) cout << "Read " << mChannel.size()
		      << " channel names from config" << endl;
    return false;
}

//======================================  Open an output writer.
bool
FrWrite::openWriter(const Time& gps, Interval dt)  {
    const string seq_esc("%q");
    if (Debug() > 2) cout << "FrWrite: in openWriter." << endl;

    //--------------------------------  Generate an output file name
    string format = mOutFile;
    if (format.find(seq_esc) != string::npos) {
        int iseq = 0;
	if (mSequence) iseq = mFileOut/mSequence;
	format.replace(format.find(seq_esc), seq_esc.size(), to_string(iseq));
    }
    
    frame_name file;
    file.split(format);
    file.make_dir(file.dir_name(gps));

    //--------------------------------  Open the output file
    mCurFile = file.file_path(gps, dt);
    unique_ptr< frame_buffer > fb(new frame_buffer(ios::out));
    const size_t buf_size = 128*1024;
    fb->pubsetbuf(new char[buf_size], buf_size);
    if (fb->open(mCurFile.c_str(), ios::out)) {
        if (Debug() > 1) cout << "FrWrite: Opened " << mCurFile
			      << " for frame output" << endl;
    } else {
        cerr << "FrWrite: Can't open " << mCurFile << " for output." << endl;
	return true;
    }

    //--------------------------------  Create a frame writer.
    try {
        mWriter.reset(new FrameCPP::OFrameStream(fb.release()));
    } catch (std::exception& e) {
        mWriter.reset();
	cerr << "FrWrite: Unable to open writer: " << e.what() << endl;
    }
    if (!mWriter) return true;

    //--------------------------------  Set counter(s), return
    mFrDone = 0;
    return false;
}

//======================================  Process one frame (FrameCPP).
void
FrWrite::ProcessFrame(frame_ptr_type frame)
{
    if (Debug() > 2) cout << "FrWrite: in ProcessFrame." << endl;

    //----------------------------------  Loop over subframes
    Interval lenF = getDacc().getDt();
    Interval dT = mResize;
    if (!dT) dT = lenF;
    for (Interval Off=0.0; Off<lenF; Off += dT) {

        //------------------------------  See if this frame passes the test
        if (mOSC) {
	    getDacc().zeroChans(dT);
	    getDacc().fillChans(Off, dT);
	    if (mOSC->satisfied(mTrigName)) {
		mBackWrite = mBacklog.size() + mTrigFrame;
	    }
	} else {
	    mBackWrite = 1;
	}

	//cout << "mBackWrite = " << mBackWrite << endl;
	//cout << "mBackValid = " << mBacklog.size() << endl;
	//cout << "mTrigFrame = " << mTrigFrame << endl;
	//cout << "mBackFrame = " << mBackFrame << endl;
    
	//------------------------------  Open a writer if necessary
	if (mBackWrite) {
 	    frame_ptr_type newframe = MakeFrame(frame, Off, dT);
	    if (!mWriter && mBackWrite) {
	        GPSTime T = (!mBacklog.empty()) ? mBacklog[0]->GetGTime() 
		            : newframe->GetGTime();
		Time t0(T.getSec(), T.getNSec());
		if (openWriter(t0, dT)) {
		    finish();
		    return;
		}
	    }

	    //--------------------------  Write the frame
	    Time tFrame;
	    if (!mBacklog.empty()) {
	        WriteFrame(mBacklog.front());
		tFrame = GetGPSTime(mBacklog.front()->GetGTime()); 
		mBacklog.erase(mBacklog.begin());
		mBacklog.push_back(newframe);
	    } else {
		WriteFrame(newframe);
		tFrame = GetGPSTime(newframe->GetGTime()); 
	    }
	    if (Debug()) cout << "FrWriter: Wrote " << dT << " second frame at "
			      << tFrame.getS() << endl;
	    mBackWrite--;
	    mFrDone++;
	    mFrTotal++;

	    //--------------------------  See if this is the end
	    if (mFrDone >= mFrOut) {
	        closeWriter();
		if (mRepeat != 0 && mFileOut >= mRepeat) finish();
	    }

	    //--------------------------  Record the frame if necessary
	} else if (mBackFrame) {
	    if (mBacklog.size() == mBackFrame) {
		mBacklog.erase(mBacklog.begin());
	    }
	    mBacklog.push_back(MakeFrame(frame, Off, dT));
	}
    }
}

//======================================  Write out the frame
//                                        Note that this kludge is needed to 
//                                        write a const FrameH with FrameCPP
//                                        version 6
void
FrWrite::WriteFrame(frame_ptr_type frame) {
    FrameSize(*frame);
    if (Debug() > 2) cerr << "Frwrite: Start writing frame." << endl;
    try {
        mWriter->WriteFrame(frame);
    } catch (std::exception& c) {
        cerr << "FrWrite: Caught exception in WriteFrame: " 
	     << c.what() << endl;
    } catch (...) {
        cerr << "FrWrite: Caught unidentified exception in WriteFrame." 
	     << endl;
    }
    if (Debug() > 2) cout << "Frwrite: Finished writing frame." << endl;    
}

//======================================  Copy selected parts of a frame;
FrWrite::frame_ptr_type
FrWrite::MakeFrame(frame_ptr_type frame, Interval Off, Interval dT)
{
    frame_ptr_type newFrame;

    if (Debug() > 2) cout << "FrWrite: starting MakeFrame." << endl;
    Time tStart;
    if (mStartTime) {
	if (!mNextTime) tStart = mStartTime;
	else            tStart = mNextTime;
    } else {
	tStart = GetGPSTime(frame->GetGTime()) + Off;
	if (mTimeOff) tStart += mTimeOff;
    }
    mNextTime = tStart;
    GPSTime gStart(tStart.getS(), tStart.getN());
    Interval dTall = frame->GetDt();

    //--------------------------------  Check for raw data in frame
    frrawdata_pointer raw(frame->GetRawData());
    if (!raw) {
        cerr << "FrWrite:ProcessFrame: No raw data in frame." << endl;
	return newFrame;
    }

    //--------------------------------  Process according to mode.
    switch (mMode) {

    //--------------------------------  Write frame unmodified.
    case m_CopyAll:
	if (dT == dTall) {
	    if (gStart != frame->GetGTime()) frame->SetGTime(gStart);
	    newFrame = frame;
	} else {

	    //----------------------------  Build a frame with identical params
	    newFrame = frame_ptr_type(new FrameH(frame->GetName(),
						 frame->GetRun(), 
						 frame->GetFrame(), gStart,
						 frame->GetULeapS(), dT));
	    newFrame->RefDetectSim()  = frame->RefDetectSim();
	    newFrame->RefDetectProc() = frame->RefDetectProc();

	    //----------------------------  Add a like-named raw data block
	    newFrame->RefHistory()  = frame->RefHistory();
	    newFrame->RefAuxData()  = frame->RefAuxData();
	    newFrame->RefProcData() = frame->RefProcData();
	    newFrame->RefSimData()  = frame->RefSimData();
	    newFrame->RefSimEvent() = frame->RefSimEvent();
	    frrawdata_pointer newraw(new FrRawData(raw->GetName()));
	    newFrame->SetRawData(newraw);
	    auto& cadc = raw->RefFirstAdc();
	    for (auto p=cadc.begin(); p != cadc.end(); p++) {
		cadc.append((*p)->Subset(Off, dT));
	    }
	}
	break;

    //--------------------------------  Select specific channels
    case m_Select:

        //----------------------------  Build a frame with identical params
        newFrame = frame_ptr_type(new FrameH(frame->GetName(), frame->GetRun(), 
					     frame->GetFrame(), gStart,
					     frame->GetULeapS(), dT));
	newFrame->RefDetectSim()  = frame->RefDetectSim();
	newFrame->RefDetectProc() = frame->RefDetectProc();

        //----------------------------  Add a like-named raw data block
	newFrame->RefHistory()  = frame->RefHistory();
	newFrame->RefAuxData()  = frame->RefAuxData();
	newFrame->RefProcData() = frame->RefProcData();
	newFrame->RefSimData()  = frame->RefSimData();
	newFrame->RefSimEvent() = frame->RefSimEvent();
	newFrame->SetRawData(frrawdata_pointer(new FrRawData(raw->GetName())));

        //----------------------------  Loop over channel list, copy channels.
	for (chan_iter itr=mChannel.begin(); itr != mChannel.end(); itr++) {
	    string ChName = itr->getName();
	    if (Debug()> 3) cout << "Get channel: " << ChName << endl;

	    //------------------------  Get the adc pointer
	    fradcdata_pointer p;
	    int b4 = getDacc().findAdcOrdered(ChName, p);
	    if (!b4) {
		if (Debug() > 3 || !itr->getError()) {
		    cerr << "Channel " << ChName << " not found in frame "
			 << getNFrames() << endl;
		}
		itr->setError(true);
		continue;
	    } else {
		if (!Off && dT == dTall) itr->Copy(p, newFrame);
		else                     itr->Copy(p, newFrame, Off, dT);
		itr->setError(false);
		if (b4 < 0 && itr != mChannel.begin()) {
		    chan_iter prev = itr; prev--;
		    if (Debug() > 3) {
			cout << "Reorder channels " << ChName
			     << " <-> " << prev->getName() << endl;
		    }
		    std::swap<ChanDesc>(*itr, *prev);
		}
	    }
	}
    }
    if (Debug() > 2) cout << "FrWrite: exit MakeFrame." << endl;
    mNextTime += dT;
    return newFrame;
}

//======================================  Print final Results
void
FrWrite::closeWriter(void)
{
    if (Debug() > 2) cout << "FrWrite: in closeWriter." << endl;
    if (mWriter) {
        if (Debug() > 1) {
	    cout << "Closing file: " << mCurFile << endl;
	}
        mWriter->Close();
        mWriter.reset();
	mFrameBuf = 0;
	mFileOut++;
    }

    if (mFrameBuf) {
        delete mFrameBuf;
	mFrameBuf = 0;
    }
}

//======================================  Print final Results
FrWrite::~FrWrite()
{
    closeWriter();
    if (mFileOut) {
        cout << "FrWriter terminating after writing " << mFileOut 
	     << " files containing " << mFrTotal << " total frames." 
	     << endl;
    }
}

//======================================  Channel descriptor constructors.
ChanDesc::ChanDesc(void) 
  : mRate(0), mError(false)
{
}

ChanDesc::ChanDesc(const char* configstr) 
  : mRate(0.0), mError(false)
{
    int lName = string(configstr).find(' ');
    if (lName <= 0) lName = strlen(configstr);
    mName = string(configstr, lName);
}

//======================================  Make a FrRawData structure
frrawdata_pointer
makeRaw(frameh_pointer out, const std::string name="Raw Data") {
    frrawdata_pointer raw(out->GetRawData());
    if (!raw) {
        out->SetRawData(frrawdata_pointer(new FrRawData(name)));
	raw = out->GetRawData();
    }
    return raw;
}

//======================================  Copy Adc data into a frame
void
ChanDesc::Copy(fradcdata_pointer in, frameh_pointer out, 
	       Interval Off, Interval dT) {

    //----------------------------------  Get the raw data block
    frrawdata_pointer raw(makeRaw(out));

    //----------------------------------  Copy the adc data for this channel
    FrRawData::firstAdc_type* cadc = &raw->RefFirstAdc();
    auto pvec(in->RefData().front());
    pvec->Uncompress();
    fradcdata_pointer newA(in->Subset(Off, dT));
    cadc->append(newA);
}

//======================================  Copy Adc data into a frame
void
ChanDesc::Copy(fradcdata_pointer in, frameh_pointer out) {

    //----------------------------------  Get the raw data block
    frrawdata_pointer raw(makeRaw(out));

    //----------------------------------  Copy the adc data for this channel
    FrRawData::firstAdc_type* cadc = &raw->RefFirstAdc();
    cadc->append(in);
}

//======================================  Copy adc from one frame to another
void
ChanDesc::Copy(frameh_pointer in, frameh_pointer out, 
	       Interval Off, Interval dT) {
    frrawdata_pointer raw(in->GetRawData());
    if (!raw) {
        cout << "No raw data in input frame" << endl;
    } else {
        FrRawData::const_firstAdc_iterator AdcIter;
	AdcIter = (raw->RefFirstAdc()).find(getName(), (raw->RefFirstAdc()).begin());
	if (AdcIter != (raw->RefFirstAdc()).end()) {
	    if (!Off && dT == in->GetDt()) Copy(*AdcIter, out);
	    else                           Copy(*AdcIter, out, Off, dT);
	    setError(false);
	} else if (!getError()) {
	    cout << "Channel: " << getName() << " is not in the input frame."
		 << endl;
	    setError(true);
	}
    }
}

//======================================  Set Chandesc error bit.
void
ChanDesc::setError(bool t) {
    mError = t;
}
